from dataclasses import dataclass
from typing import List
from abc import abstractmethod, ABCMeta


class Element(metaclass=ABCMeta):

    @abstractmethod
    def to_markdown(self) -> str:
        pass


@dataclass
class Link(Element):
    text: str
    href: str

    def to_markdown(self) -> str:
        return f"[{self.text}]({self.href})"


@dataclass
class Image(Element):
    alt: str
    src: str

    def to_markdown(self) -> str:
        return f"![{self.alt}]({self.src})"


class Li(Element):
    elements: List[Element] = []

    def __init__(self, elements: List[Element] = []) -> None:
        self.elements = elements

    def add(self, element: Element) -> None:
        self.elements.append(element)

    def to_markdown(self) -> None:
        pass


class ULi(Li):
    def to_markdown(self) -> str:
        from ..generator import md
        return "\n".join([f"-\t{md(el)}" for el in self.elements])


class OLi(Li):
    def to_markdown(self) -> str:
        from ..generator import md
        return "\n".join([f"{i+1}.\t{md(self.elements[i])}" for i in range(len(self.elements))])


@dataclass
class P(Element):
    text: str = ""

    def to_markdown(self) -> str:
        return self.text


@dataclass
class Header(Element):
    size: int
    text: str

    def to_markdown(self) -> str:
        return f"{'#'*self.size} {self.text}"


class Br(Element):
    def to_markdown(self) -> str:
        return "\n"


class Hr(Element):
    def to_markdown(self) -> str:
        return "---"


class Table(Element):
    data: List[dict]
    keys: List[str]
    capitalized_headers: bool = True

    def __init__(self, data: List[dict], keys: List[str], capitalized_headers: bool = True) -> None:
        self.data = data
        self.keys = keys
        self.capitalized_headers = capitalized_headers

    def to_markdown(self) -> str:
        markdown_text = ""
        markdown_text += "| "+" | ".join([key.capitalize() if self.capitalized_headers else key for key in self.keys]) + " |\n"
        markdown_text += '| ----------- '*len(self.keys)+"|\n"
        for d in self.data:
            markdown_text += "| "+" | ".join([d.get(key, "") for key in self.keys]) + " |\n"
        return markdown_text.strip("\n")


@dataclass
class CodeBlock(Element):
    lang: str
    code: str

    def to_markdown(self) -> str:
        from ..generator import md
        return f"""```{self.lang}\n{self.code.strip()}\n```"""


class TaskList(Element):
    @dataclass
    class Task:
        text: str
        checked: bool = False

    tasks: List[Task]

    def __init__(self, tasks: List[Task]) -> None:
        self.tasks = tasks

    def to_markdown(self) -> str:
        return "\n".join([f"- [{'x' if task.checked else ' '}] {task.text}" for task in self.tasks])


@dataclass
class BlockQuote(Element):
    text: str

    def to_markdown(self) -> str:
        return f"> {self.text}"


@dataclass
class Strikethrough(Element):
    text: str

    def to_markdown(self) -> str:
        return f"~~{self.text}~~"


class DefinitionList(Element):
    text: str
    items: List[Element]

    def __init__(self, text: str, items: List[Element]) -> None:
        self.text = text
        self.items = items

    def to_markdown(self) -> str:
        from ..generator import md
        return "\n".join([self.text, *[f":\t{md(item)}" for item in self.items]])

@dataclass
class Footnote(Element):
    text: str
    definition: Element

    def to_markdown(self) -> str:
        from .utils import gen_random
        from ..generator import md
        random_smth = gen_random()
        return f"{self.text} [^{random_smth}]\n [^{random_smth}]: {md(self.definition)}"



from ..elements import Element, Link, Header, Br
from typing import List, Any


class Section:
    title: str = ""
    inner: List[Element] = []
    header_size: int = 2
    have_header: bool = True

    def to_markdown(self):
        from ..generator import md
        markdown_text: str = ""
        markdown_text += (md(Header(self.header_size, self.title)) + md(Br())*2) if self.have_header else ""

        for el in self.inner:
            markdown_text += md(el) + md(Br())*2

        return markdown_text


class Readme:
    title: str = ""
    description: str = ""
    badges: List[Any] = []
    sections: List[Section] = []
    have_header: bool = True

    def __init__(self, title: str = "", description: str = "", sections: List[Section] = [], badges: List[Any] = [], have_header: bool = True) -> None:
        self.title = title
        self.description = description
        self.badges = badges
        self.sections = sections
        self.have_header = have_header

    def to_markdown(self):
        from ..generator import md
        markdown_text: str = ""
        markdown_text += (md(Header(1, self.title)) + md(Br())*2) if self.have_header else ""

        for section in self.sections:
            markdown_text += md(section)

        return markdown_text


class LicenseSection(Section):
    def __init__(self, license_type: str, license_link: str) -> None:
        self.title = "License"
        self.inner.append(Link(license_type, license_link))



from .elements import (
    Element,
    Link,
    Image,
    ULi,
    OLi,
    P,
    Header,
    Br,
    Table,
    CodeBlock, 
    Footnote, 
    TaskList
)
from .styles import (
    bold,
    italic,
    code
)


from ..structure import Readme,Section
from ..elements import Element
from typing import Union

def generate(readme_obj: Readme, type: str='md', file_name:str="README.md"):
    inner_data:str = ""
    
    if type == 'md':
        inner_data = readme_obj.to_markdown()
    
    with open(file_name,"w") as f:
        f.write(inner_data)

def md(element:Union[Element,Section]):
    return element.to_markdown()